'''
Comprehensive construction of empCpds via generic tree structures.
Each empCpd = annoTree.
1. construct all isotopologues into a list of trees
2. attach in-src modifications (adducts) to nodes in each tree
3. for peaks without isotopologues, add to new trees and attach adducts

This applies to regular LC-MS data, but also
enables easy analysis of isotope tracing and chemical labeling data.

Example data:

    >>> node2tree = construct_isotopic_trees(plist)
    Found 873 isotopic pairs, 703 trees and 170 in branches.
    >>> node2tree['F795'].root
    'F649'
    >>> node2tree['F795'].nodes
    {'F649': Node(tag=179.1123@172.9, identifier=F649, data=None), 'F755': Node(tag=180.1156@172.9, identifier=F755, data=None), 
    'F794': Node(tag=181.1198@171.7, identifier=F794, data=None), 'F795': Node(tag=181.1198@173.3, identifier=F795, data=None), 
    'F839': Node(tag=182.1233@171.7, identifier=F839, data=None)}
    >>> node2tree['F795'].show()
    179.1123@172.9
    └── 180.1156@172.9
        ├── 181.1198@171.7
        │   └── 182.1233@171.7
        └── 181.1198@173.3


Trees use peak IDs, but full peak data are in json peak lists.

Shuzhao 2022-09-22

#
# to-dos: Above only includes isotopic peaks; will add peaks without C13 counterparts into empCpd trees
#
# 

The trees can be annotated by customized methods, e.g.
- in-house compound library
- targeted pathway search
- public databases
- linking tandem mass spec data


'''

import os
import json
# from itertools import combinations
import treelib

from .search import build_centurion_tree, find_all_matches_centurion_indexed_list


def make_peak_tag(peak):
    '''
    peak format: {'id': 'F1', 'mz': 60.0808, 'rtime': 117.7, 'intensities': [250346.0], 'representative_intensity': 250346.0}
    '''
    return str(round(peak['mz'], 4)) + '@' + str(round(peak['rtime'], 1))

def make_peak_dict(peak_list):
    peak_dict = {}
    for p in peak_list:
        peak_dict[p['id']] = p
    return peak_dict

def construct_isotopic_trees(peak_list, 
                    search_patterns = [(1.003355, '13C/12C', (0, 0.8))], 
                    mz_tolerance_ppm=5, 
                    isotope_rt_tolerance=2, 
                    check_isotope_ratio = False,
                    tree_depth_limit=10):
    '''
    Make isotopic trees from peak_list. 
    tree_depth_limit: limit of tree depth, i.e. steps on a branch.
    For other parameters, see get_isotopic_pairs.
    Return node2tree, a dict of peak to tree mapping.
    No peak is assigned to more than one trees. But some close peaks can't be distinguished here.

    Example
    =======
    >>> node2tree['F515'].show()
    173.0922@172.9
    └── 174.0956@171.5
        ├── 175.0991@172.4
        │   ├── 176.1031@171.7
        │   │   ├── 177.1065@171.7
        │   │   └── 177.1065@174.1
        │   └── 176.1031@174.1
        └── 175.0991@173.1

    >>> node2tree['F796'].show()
    180.1244@171.2
    └── 181.1279@170.1
        └── 182.1313@170.1
    '''
    peak_dict = make_peak_dict(peak_list)
    mztree = build_centurion_tree(peak_list)
    isotopologues = get_isotopic_pairs(peak_list, mztree, search_patterns, mz_tolerance_ppm, 
                    isotope_rt_tolerance, check_isotope_ratio)
    # isotopologues format: [ ((195, 'anchor'), (206, '13C/12C')), ...]
    # build trees
    annoTrees, branches = [], []
    all_target_nodes = set([x[1][0] for x in isotopologues])
    for pair in isotopologues:
        if pair[0][0] not in all_target_nodes:   # if source_node appears in any of the target_nodes, it's not a root
            tree = treelib.Tree()
            tree.create_node(make_peak_tag(peak_dict[pair[0][0]]), pair[0][0], data=pair[0][1])
            tree.create_node(make_peak_tag(peak_dict[pair[1][0]]), pair[1][0], parent=pair[0][0], data=pair[1][1])
            annoTrees.append(tree)
        else:
            branches.append(pair)
    
    print("Found %d isotopic pairs, %d trees and %d in branches." %(len(isotopologues), len(annoTrees), len(branches)))
    
    node2tree = {}
    for tree in annoTrees:
        for node in tree.nodes:
            node2tree[node] = tree

    # do branches now
    remaining = []
    for pair in branches:          
        if pair[0][0] in node2tree:                  # pair[0] already in a tree
            try:
                this_tree = node2tree[pair[0][0]]
                this_tree.create_node( make_peak_tag(peak_dict[pair[1][0]]), pair[1][0], parent=pair[0][0], data=pair[1][1] )
                node2tree[pair[1][0]] = this_tree
            except treelib.exceptions.DuplicatedNodeIdError:
                # print("already included ", pair)
                pass                 # pair already in a tree
        else:
            remaining.append(pair)

    steps = 0
    while remaining and steps < tree_depth_limit:
        tmp = []
        for pair in remaining:          
            if pair[0][0] in node2tree: 
                try:
                    this_tree = node2tree[pair[0][0]]
                    this_tree.create_node( make_peak_tag(peak_dict[pair[1][0]]), pair[1][0], parent=pair[0][0], data=pair[1][1] )
                    node2tree[pair[1][0]] = this_tree
                except treelib.exceptions.DuplicatedNodeIdError:
                    # print("already included ", pair)
                    pass               # pair already in a tree
            else:
                tmp.append(pair)
        remaining = tmp
        steps += 1

    return node2tree

def get_isotopic_pairs(list_peaks, 
                    mztree, 
                    search_patterns = [(1.003355, '13C/12C', (0, 0.8))],
                    mz_tolerance_ppm=5, 
                    isotope_rt_tolerance=2, 
                    check_isotope_ratio = False,
                    ):
    '''
    To find all isotope pairs. Similar to search.get_seed_empCpd_signatures, but return unidirectional pairs only.
    If input peaks have overlaps/duplicates, result will contain the redundant overlap peaks.

    Input
    =====
    list_peaks: [{'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'rtime': 654, 'height': 14388.0, 'id': 555}, ...]
    mztree: indexed list_peaks
    mz_tolerance_ppm: ppm tolerance in examining m/z patterns.
    seed_search_patterns: a list in the format of [(mz difference, notion, (ratio low limit, ratio high limit)), ..]
            This can be obtained through search.isotopic_patterns. The ratios are optional, because 
            1) naturally occuring constrains are based on chemical formula;
            2) rules are different when isotope tracers are introduced to the experiments.
    isotope_rt_tolerance: tolerance threshold for deviation in retetion time, arbitrary unit depending on input data.

    Return
    ======
    list of lists of peak pairs that match search_patterns patterns, e.g.
    [ ((195, 'anchor'), (206, '13C/12C')), ...]. 
    '''
    signatures = []
    for P1 in list_peaks:
        matched = [  ] 
        for _pair in search_patterns:
            (mass_difference, relation) = _pair[:2]
            tmp = find_all_matches_centurion_indexed_list(P1['mz'] + mass_difference, mztree, mz_tolerance_ppm)
            for P2 in tmp:
                if abs(P1['rtime']-P2['rtime']) <= isotope_rt_tolerance:
                    if check_isotope_ratio and len(_pair) > 2:  # checking abundance ratio
                        (abundance_ratio_min, abundance_ratio_max) = _pair[2]
                        if abundance_ratio_min*P1['height'] < P2['height'] < abundance_ratio_max*P1['height']:
                            matched.append( ((P1['id'], 'anchor'), (P2['id'], relation)) )
                    else:
                        matched.append( ((P1['id'], 'anchor'), (P2['id'], relation)) )
 
        signatures += matched

    return signatures


def rt_matched_by_tolerance(P1, P2, rt_tolerance):
    return abs(P2['rtime']-P1['rtime']) < rt_tolerance

def rt_compared_by_values(P1, P2, rt_tolerance=None):
    return P2['rtime'] > P1['rtime']

def merge_trees_by_modifications(trees, list_peaks, search_patterns, rt_verify_function,
                    mz_tolerance_ppm=5, rt_tolerance=10):
    '''
    Merge trees that are from same compound but with  modifications. 
    rt_verify_function applies rules on retention time.
    '''
    peak_dict = make_peak_dict(list_peaks)
    tree_dict = {}
    for tree in trees: tree_dict[tree.root] = tree
    print("Merging adducts on %d trees..." %len(trees))
    matched_pairs = []
    relevant_peak_centTree = build_centurion_tree( [peak_dict[p] for p in tree_dict.keys()] )
    for p in tree_dict:
        P1 = peak_dict[p]
        matched = [  ] 
        for _pair in search_patterns:
            (mass_difference, relation) = _pair[:2]
            tmp = find_all_matches_centurion_indexed_list(P1['mz'] + mass_difference, 
                                                    relevant_peak_centTree, mz_tolerance_ppm)
            for P2 in tmp:
                if rt_verify_function(P1, P2, rt_tolerance):
                    matched.append( (P1['id'], P2['id'], relation) )
        matched_pairs += matched

    list_primary = set([x[0] for x in matched_pairs])
    list_adducts = set([x[1] for x in matched_pairs])
    overlap = list_primary.intersection(list_adducts)
    union = list_primary.union(list_adducts)
    if overlap: print("Unresolved multiple relationships: ", overlap)

    good_trees = [tree_dict[x] for x in tree_dict if x not in union]
    for P in matched_pairs:
        p1, p2 = P[:2]
        tree_dict[p2].get_node(p2).data = P[2]      # add relation note
        try:
            tree_dict[p1].paste(p1, tree_dict[p2])
        except ValueError:                  # likely Duplicated nodes
            # print(tree_dict[p1].show(), tree_dict[p2].show())
            for node in tree_dict[p2].nodes:
                if node not in tree_dict[p1].nodes:
                    tree_dict[p1].add_node(tree_dict[p2].nodes[node], parent=tree_dict[p1].root)
        good_trees.append(tree_dict[p1])

    print("Got %d merged trees." %len(good_trees))
    return good_trees


def merge_trees_by_insrc_modifications(trees, list_peaks,
                    search_patterns = [(1.0078, 'H'), (21.9820, 'Na/H'), (41.026549, 'Acetonitrile')], 
                    mz_tolerance_ppm=5, rt_tolerance=10):
    '''
    Merge trees that are from same compound but with diff adducts or neutral loss,
    with user supplied search_patterns (can be from search.common_adducts).
    These in-source modifications must have same retention time.
    '''
    return merge_trees_by_modifications(trees, list_peaks,
                    search_patterns, rt_verify_function=rt_matched_by_tolerance,
                    mz_tolerance_ppm=mz_tolerance_ppm, rt_tolerance=rt_tolerance)

def merge_trees_by_derivatization(trees, list_peaks,
                    search_patterns = [(161.08407, 'DmPA'), (233.05105, 'Dens'), (247.07793, 'DnsHz')], 
                    mz_tolerance_ppm=5, ):
    '''
    Merge trees that are from same compound but with chemical labeling/derivatization,
    which leads to greater retention time.
    '''
    return merge_trees_by_modifications(trees, list_peaks,
                    search_patterns, rt_verify_function=rt_compared_by_values,
                    mz_tolerance_ppm=mz_tolerance_ppm, rt_tolerance=None)


def export_json_trees(trees, outfile="export_annoTree.tsv"):
    pass


def export_tsv_trees(trees, outfile="export_annoTree.tsv"):
    s = 'Feature_ID\tFeature_tag\troot\troot_tag\trelation\n'
    for tree in trees:
        for node in tree.nodes:
            ND, ROOT = tree.get_node(node), tree.get_node(tree.root)
            s += '\t'.join([ND.identifier, ND.tag, ROOT.identifier, ROOT.tag, ND.data or '']) + '\n'
    with open(outfile, 'w') as O:
        O.write(s)




# ------------------------------------------------------------

class empTree:
    '''
    Not used now. 
    Container for empirical tree, as ambiguous branches don't fit a strict tree structure.  

    Trees break down here when leaves are shared:
        118.0652@109.9
        ├── 119.0685@109.2
        └── 159.0917@109.9
            └── 160.0952@110.4

        159.0917@110.4
        └── 160.0952@110.4

    '''
    def __init__(self) -> None:
        self.ambiguous = False
        self.uniqe_nodes = []
        self.trees = []





'''
Handling chemical formulae.

This should switch to a more dedicated chemical tool, 
handling structural formats too.

Primary goal in the Mummichog suite is to compute:
1. associations btw chemical formulae and masses
2. Isotopes
3. In source reactions, including adducts, neutral loss, fragments
4. Biochemical reactions

'''

import re
from collections import namedtuple
import warnings

Ion = namedtuple('Ion', ['mz', 'ion', 'delta_formula'])

PROTON = 1.00727646677
electron = 0.000549

# prepared by Minghao Gong
# from: https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-relative-atomic-masses
# Using mass of most abundant isotope.
atom_mass_dict = {
    'C': 12.0000,
    'H': 1.0078250322,
    'O': 15.99491461957,
    'N': 14.00307400443,
    'S': 31.9720711744,
    'P': 30.97376199842,
    'Fe': 55.93493633,
    'He': 4.00260325413,
    'Li': 7.0160034366,
    'Be': 9.012183065,
    'B': 11.00930536,
    'F': 18.99840316273,
    'Ne': 19.9924401762,
    'Na': 22.9897692820,
    'Mg': 23.985041697,
    'Al': 26.98153853,
    'Si': 27.97692653465,
    'Cl': 34.968852682,
    'Ar': 39.9623831237,
    'K': 38.9637064864,
    'Ca': 39.962590863,
    'Sc': 44.95590828,
    'Ti': 47.94794198,
    'V': 50.94395704,
    'Cr': 51.94050623,
    'Mn': 54.93804391,
    'Co': 58.93319429,
    'Ni': 57.93534241,
    'Cu': 62.92959772,
    'Zn': 63.92914201,
    'Ga': 68.9255735,
    'Ge': 73.921177761,
    'As': 74.92159457,
    'Se': 79.9165218,
    'Br': 83.9114977282,
    'Kr': 77.92036494,
    'Rb': 84.9117897379,
    'Sr': 87.9056125,
    'Y': 88.9058403,
    'Zr': 89.9046977,
    'Nb': 92.9063730,
    'Mo': 97.90540482,
    'Tc': 96.9063667,
    'Ru': 101.9043441,
    'Rh': 102.9054980,
    'Pd': 105.9034804,
    'Ag': 106.9050916,
    'Cd': 113.90336509,
    'In': 114.903878776,
    'Sn': 117.90160657,
    'Sb': 120.9038120,
    'Te': 129.906222748,
    'I': 126.9044719,
    'Xe': 131.9041550856,
    'Hg': 201.97064340,
    'Pb': 207.9766525,
    'Ba': 137.905247,

    # Common isotopic mass
    '[13C]': 13.00335483507,
    '[15N]': 15.00010889888,
    'D':  2.01410177812,
    '[37Cl]':  36.965902602,
    '[18O]': 17.99915961286,
    '[34S]': 33.967867004,
    '[81Br]': 80.916291,
    '[56Fe]': 55.934938,
    '[7Li]': 7.016005,
    '[11B]': 11.009305,
    '[65Cu]': 64.927794,
    '[109Ag]': 108.904756
    }


def parse_chemformula_dict(x):  # return to its original form
    '''This does not deal with nested groups in chemical formula, or isotopes.
    Formula from HMDB 3.5 are compatible.
    Example
    -------
    parse_chemformula_dict('C3H6N2O2')
    {'C': 3, 'H': 6, 'N': 2, 'O': 2}
    Note
    ----
    If formula contains negative number of elements, it will have no warnings but get a errenous dictionary (e.g., 'C2H12O-5' will return {'C': 2, 'H': 12, 'O': 1})
    Be careful of those circumstances.
    '''
    p = re.findall(r'([A-Z][a-z]*)(\d*)', x)
    d = {}
    for pair in p: d[pair[0]] = int( pair[1] or 1 )
    return d



def parse_chemformula_dict_comprehensive(x, remove_additive_cpd = True, handle_negative_number = True):
    '''
    A more inclusive parsing function (compare to `parse_chemformula_dict`) to handle some rare cases of formula parsing
    This also deals with nested groups in chemical formula. And the number of repetitive elements present in the formula will be sumed (e.g., CH2O2H24O4)
    Default: removing additive compounds (e.g., [13C]4H12N2·2HCl will turn into [13C]4H12N2)
    Default: Handle negative number in parsing dictionary (e.g.,, C-2H-2O-2)
    Handles common isotopes.
    Formula from HMDB 3.5 are compatible.

    Example
    -------
    parse_chemformula_dict_comprehensive('C-2H12O5H-4·H2O', remove_additive_cpd=True, handle_negative_number=True)
    {'C': -2, 'H': 8, 'O': 5}
    '''
    if remove_additive_cpd == True: # remove additive compound is true
        if '·' in x:
            x = x.split('·')[0]
    if handle_negative_number == True: # handle formula which has negative value; If the formula somehow has `-` not indicating negative number, this function will not work
        p = re.findall(r'(\[\d*[A-Z][a-z]*\]|[A-Z][a-z]*)(-?\d*)', x)
    else: 
        p = re.findall(r'([A-Z][a-z]*)(\d*)', x)
    k = set([x[0] for x in p]) # handle repetitive element (nested group)
    d = {k:int(0) for k in k}
    for pair in p: d[pair[0]] += int( pair[1] or 1 )
    return d


def add_formula_dict(dict1, dict2):
    '''
    Addition of two formulae as dictionaries.
    This allows calculating formula after a reaction, as dict2 can contain substraction of elements.
    Not as good as using real chemical structures, just a simple approximation.

    '''
    new, result = {}, {}
    for k in set(dict1.keys()).union( set(dict2.keys()) ):
        if k in dict1 and k in dict2:
            new[k] = dict1[k] + dict2[k]
        elif k in dict1:
            new[k] = dict1[k]
        else:
            new[k] = dict2[k]
    # check validity, as no negative number of element is allow in product
    _VALID = True
    for k in new:
        if new[k] < 0:
            _VALID = False
        elif new[k] > 0:     # remove if element is 0
            result[k] = new[k] 
    if _VALID:
        return result
    else:
        return None


def calculate_formula_mass(formula, massDict=atom_mass_dict):
    '''
    Modified from MG's calculate_mass. Rounding is not useful during calculation. 
    '''
    formula_dict = parse_chemformula_dict(formula)
    _m = 0
    for k,v in formula_dict.items():
        _m += v * massDict[k]
    return _m


def calculate_mass(formula_dict, decimal_places=6, massDict=atom_mass_dict):
    '''
    Calculate mass from the formula dictionary, with rounded decimals. 
    Modified for MG. 
    '''
    _m = 0
    for k,v in formula_dict.items():
        _m += v * massDict[k]
    _m = round(_m,decimal_places)
    return _m
    

def check_elemental_subset(Fm1, Fm2):
    '''
    Check if Fm2 a subunit of Fm1; Dictionary representation of chem formula.
    Not used now.
    '''
    Fm2_in_Fm1 = True
    Elements1 = Fm1.keys()
    Elements2 = Fm2.keys()
    if [x for x in Elements2 if x not in Elements1]:    #element not in Fm1
        return False
    else:
        for e in Elements2:
            if Fm2[e] > Fm1[e]: Fm2_in_Fm1 = False
        return Fm2_in_Fm1


def dict_to_hill_formula(d):
    '''
    Generate simple formula from dictionary.
    Hill notion: First C, H and other elements in alphabetical order.
    Carbon isotope (C13) is considered, but not other isotopic elements for now.

    Input 
    -----
    formula dictionary, e.g. {'C': 3, 'H': 6, 'N': 2, 'O': 2} 

    Return
    ------
    Formula in Hill notion, e.g. 'C3H6N2O2'

    '''
    def _str_element_number(int_x):
        # convert int_x to str and '' for int_x is 1. Not checking validity. 
        if int_x == 1:
            return ''
        else:
            return str(int_x)

    elements = list(d.keys())
    elements.sort()
    new_order = [x for x in elements if x not in ['C', 'H', '(C13)']]
    if 'H' in d:
        new_order = ['H'] + new_order
    if '(C13)' in d:
        new_order = ['(C13)'] + new_order
    if 'C' in d:
        new_order = ['C'] + new_order

    formula = ''
    for x in new_order:
        formula += x + _str_element_number(d[x])
    return formula


def __get_adduct_list__(mw, mode, primary_only):
    '''
    Return list of adduct or isotope, [(mz, ion, formula_change_dictionary), ...].
    Next step compute_adducts_formulae will validate the ion.
    Enforcing primary ions because others are not valid without them;
    but logic implemented in search functions, not here.
    '''

    if mode == 'pos': 
        primaryList = [
            (mw - electron, 'M[1+]', {}),                              # e not changing formula
            (mw + PROTON, 'M+H[1+]', {'H': 1}),                        #  
            (mw + 21.9820 + PROTON, 'M+Na[1+]', {'Na': 1}),            # Na = 21.9820 + PROTON = 22.9893
            (mw + 18.0106 + PROTON, 'M+H2O+H[1+]', {'H': 3, 'O':1}),   #  
            (mw + 18.033823, 'M+NH4[1+]', {'H': 4, 'N': 1}),
            ]
        if primary_only:
            return [Ion(*L) for L in primaryList]

        else:
            extendedList = [
                # Use {'C':-1} to force check that C exists in formula
                (mw +1.0034 - electron, 'M(C13)[1+]', {'C':-1, '(C13)':1}), 
                (mw +1.0034 + PROTON, 'M(C13)+H[1+]', {'C':-1, '(C13)':1, 'H':1}),
                
                (mw/2 + PROTON, 'M+2H[2+]', {'H': 2}),
                (mw/3 + PROTON, 'M+3H[3+]', {'H': 3}),

                (mw/2 + 0.5017 + PROTON, 'M(C13)+2H[2+]', {'C':-1, '(C13)':1, 'H':2}),
                (mw/3 + 0.3344 + PROTON, 'M(C13)+3H[3+]', {'C':-1, '(C13)':1, 'H':3}),
                # Sulfur:  32S (95.02%), 33S (0.75%), 34S (4.21%), .
                (mw +1.9958 + PROTON, 'M(S34)+H[1+]', {'S':-1, '(S34)':1, 'H':1}),
                # Chlorine has two stable isotopes, 35Cl (75.77%) and 37Cl (24.23%)
                (mw +1.9972 + PROTON, 'M(Cl37)+H[1+]', {'Cl':-1, '(Cl37)':1, 'H':1}),
                
                (mw/2 + 10.991 + PROTON, 'M+H+Na[2+]', {'H':1, 'Na':1}),
                (mw + 37.9555 + PROTON, 'M+K[1+]', {'K':1}),         # K = 37.9555 + PROTON = 38.9628
                (mw + 57.9586 + PROTON, 'M+NaCl[1+]', {'Na': 1, 'Cl': 1}), 
                
                (mw - 18.0106 + PROTON, 'M-H2O+H[1+]', {'H': -1, 'O': -1}), 
                (mw - 36.0212 + PROTON, 'M-H4O2+H[1+]', {'H': -3, 'O': -2}),
                (mw - 17.0265 + PROTON, 'M-NH3+H[1+]', {'H': -2, 'N': -1}),
                (mw - 27.9950 + PROTON, 'M-CO+H[1+]', {'H': 1, 'C': -1, 'O': -1}),
                (mw - 43.9898 + PROTON, 'M-CO2+H[1+]', {'H': 1, 'C': -1, 'O': -2}),
                (mw - 46.0054 + PROTON, 'M-HCOOH+H[1+]', {'H': -1, 'C': -1, 'O': -2}),
                (mw + 67.9874 + PROTON, 'M+HCOONa[1+]', {'H': 1, 'C': 1, 'O': 2, 'Na': 1}),
                (mw - 72.0211 + PROTON, 'M-C3H4O2+H[1+]', {'H': -3, 'C': -3, 'O': -2}),
                (mw + 83.9613 + PROTON, 'M+HCOOK[1+]', {'H': 1, 'C': 1, 'O': 2, 'K': 1}),

                # removed, too infrequent
                # (mw - 67.9874 + PROTON, 'M-HCOONa+H[1+]', 'HCO2Na'),
                # (mw - 83.9613 + PROTON, 'M-HCOOK+H[1+]', 'HCO2K'),
                ]
            return [Ion(*L) for L in primaryList + extendedList]

    elif mode == 'neg':
        primaryList = [
            (mw - PROTON, 'M-H[-]', {'H': -1}),       
            (mw + electron, 'M[-]', {}), 
            (mw - 18.0106 - PROTON, 'M-H2O-H[-]', {'H': -3, 'O':-1}),   
            (mw + 34.9689, 'M+Cl[-]', {'Cl':1}),
            ]
        if primary_only:
            return [Ion(*L) for L in primaryList]

        else:
            extendedList = [
                (mw + 1.0034 - PROTON, 'M(C13)-H[-]', {'C':-1, '(C13)':1, 'H':-1}),
                (mw + 36.9659, 'M+Cl37[-]', {'(Cl37)':1}),
                (mw + 1.9972 - PROTON, 'M(Cl37)-H[-]', {'Cl':-1, '(Cl37)':1, 'H':-1}),
                (mw + 1.9958 - PROTON, 'M(S34)-H[-]', {'S':-1, '(S34)':1, 'H':-1}),

                (mw/2 - PROTON, 'M-2H[2-]', {'H': -2}),
                (mw + 22.9893 - 2*PROTON, 'M+Na-2H[-]', {'H':-2, 'Na':1}),
                (mw + 38.9628 - 2*PROTON, 'M+K-2H[-]', {'H':-2, 'K':1}),
                
                (mw + 78.9183, 'M+Br[-]', {'Br':1}),
                (mw + 80.9163, 'M+Br81[-]', {'(Br81)':1}),

                (mw + 40.01926853323, 'M+ACN-H[-]', {'C':2, 'H':2, 'N':1}),
                (mw + 1.007825 + 12 + 2*15.99491, 'M+HCOO[-]', {'C':1, 'H':1, 'O':2}),
                (mw + 59.013295, 'M+CH3COO[-]', {'C':2, 'H':3, 'O':2}),
                (mw - PROTON + 15.99491, 'M-H+O[-]', {'H': -1, 'O':1}),
                ]
            return [Ion(*L) for L in primaryList + extendedList]


def compute_adducts_formulae(mw, neutral_formula,  mode='pos', primary_only=False):
    '''
    Calculating isotopes and adducts, return m/z and formulae.

    A dictionary of element changes is used in the calculation.
    Another function add_formula_dict checks validity of required numbers of atoms.
    Electron is not considered in formula calculation, but may affected mass calculation.

    The isotopes, neutral losses and adducts can have multiplications,
    but multiple rounds of search will take care of those.

    Reaction notion in dictionary: 
    {'C': 3, 'H': 6, 'N': 2, 'O': 2} changed by {'C': -1, 'H': 1, 'O': -1} 
        = {'C': 2, 'H': 5, 'N': 2, 'O': 1}

    Return
    -------
    List of adducts: e.g. [(58.53894096677, 'M+2H[2+]', result_formula), ...,]

    '''
    dict_neutral_formula = parse_chemformula_dict(neutral_formula)
    adducts2get = []
    for x in __get_adduct_list__(mw, mode, primary_only):
        # ['mz', 'ion', 'delta_formula']
        result = add_formula_dict(dict_neutral_formula, x.delta_formula)
        if result:
            adducts2get.append( [x.mz, x.ion, dict_to_hill_formula(result)] )

    return adducts2get


def calculate_formula_diff(FM_D1, FM_D2):
    '''
    Calculate the differences between two formula dictionaries (FM_D1 - FM_D2), and return a formula dictionary documenting the differences
    
    '''
    unique_elements = set(list(FM_D1.keys()) + list(FM_D2.keys()))
    
    for e in unique_elements:      # update the dictionary with keys that present in another formula
        if e not in FM_D1:
            FM_D1.update({e:0})
        if e not in FM_D2:
            FM_D2.update({e:0})
    
    diff_dict = {key: FM_D1[key] - FM_D2.get(key, 0) for key in FM_D1}
    diff_dict = {k:v for k,v in diff_dict.items() if v !=0} # remove those with zero
    if not all([x > 0 for x in diff_dict.values()]) or not all([x < 0 for x in diff_dict.values()]):
        warnings.warn('the differences between the two formulas are not all positive or all negative values')
    
    return(diff_dict)

def adjust_salt_formula(F, export_str = True):
    '''
    Formula in the commercially provided products usually contain salts, but ions detected mass spec are not in salt form.
    To convert to compatible format for adduct calculation, salts (e.g., K, Na) need to replace by Hydrogens.
    '''
    exchange_dict = {
    'K': {'H': 1},
    'Na': {'H': 1},
    'Ca': {'H': 2},
    'Mg': {'H': 2}}
    F_D = parse_chemformula_dict(F, remove_additive_cpd = True)
    for k,v in exchange_dict.items():
        if k in F_D:
            for i in range(F_D[k]):
                F_D = add_formula_dict(F_D,v)
            F_D.pop(k,None)
    if export_str:
        res = dict_to_hill_formula(F_D)
    else:
        res = F_D
    return res